#!/usr/bin/env python

# Tinta
# Copyright 2024 github.com/brandoncript

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# If you use this software, you must also agree under the terms of the Hippocratic License 3.0 to not use this software in a way that directly or indirectly causes harm. You can find the full text of the license at https://firstdonoharm.dev.

"""This discover module is used to discover the colors available on your system.

Borrowing design from https://github.com/fidian/ansi

    Standard:    0   1   2   3   4   5   6   7
    Bright:      8   9  10  11  12  13  14  15

        16   17   18   19   20   21      34   35   36   37   38   39
        52   53   54   55   56   57      70   71   72   73   74   75
        88   89   90   91   92   93     106  107  108  109  110  111
        124  125  126  127  128  129     142  143  144  145  146  147
        160  161  162  163  164  165     178  179  180  181  182  183
        196  197  198  199  200  201     214  215  216  217  218  219

        22   23   24   25   26   27      40   41   42   43   44   45
        58   59   60   61   62   63      76   77   78   79   80   81
        94   95   96   97   98   99     112  113  114  115  116  117
        130  131  132  133  134  135     148  149  150  151  152  153
        166  167  168  169  170  171     184  185  186  187  188  189
        202  203  204  205  206  207     220  221  222  223  224  225

        28   29   30   31   32   33      46   47   48   49   50   51
        64   65   66   67   68   69      82   83   84   85   86   87
        100  101  102  103  104  105     118  119  120  121  122  123
        136  137  138  139  140  141     154  155  156  157  158  159
        172  173  174  175  176  177     190  191  192  193  194  195
        208  209  210  211  212  213     226  227  228  229  230  231

    Greyscale:  232   233   234   235   236   237   238   239   240   241   242   243
                244   245   246   247   248   249   250   251   252   253   254   255
"""

from typing import cast, List

from .multi_version_imports import Literal
from .utils import flatmap

STANDARD: List[int] = list(range(8))
BRIGHT: List[int] = list(range(8, 16))
DARK_GREYSCALE: List[int] = list(range(232, 244))
DARK_COLOR1: List[List[int]] = [
    [16, 17, 18, 19, 20, 21],
    [52, 53, 54, 55, 56, 57],
    [88, 89, 90, 91, 92, 93],
    [124, 125, 126, 127, 128, 129],
    [160, 161, 162, 163, 164, 165],
    [196, 197, 198, 199, 200, 201],
]

DARK_COLOR2: List[List[int]] = [
    [22, 23, 24, 25, 26, 27],
    [58, 59, 60, 61, 62, 63],
    [94, 95, 96, 97, 98, 99],
    [130, 131, 132, 133, 134, 135],
    [166, 167, 168, 169, 170, 171],
    [202, 203, 204, 205, 206, 207],
]

DARK_COLOR3: List[List[int]] = [
    [28, 29, 30, 31, 32, 33],
    [64, 65, 66, 67, 68, 69],
    [100, 101, 102, 103, 104, 105],
    [136, 137, 138, 139, 140, 141],
    [172, 173, 174, 175, 176, 177],
    [208, 209, 210, 211, 212, 213],
]
DARK_COLORS = sorted(
    flatmap(DARK_COLOR1)
    + flatmap(DARK_COLOR2)
    + flatmap(DARK_COLOR3)
    + DARK_GREYSCALE
    + STANDARD
)

LIGHT_GREYSCALE: List[int] = list(range(244, 256))
LIGHT_COLOR1: List[List[int]] = [
    [34, 35, 36, 37, 38, 39],
    [70, 71, 72, 73, 74, 75],
    [106, 107, 108, 109, 110, 111],
    [142, 143, 144, 145, 146, 147],
    [178, 179, 180, 181, 182, 183],
    [214, 215, 216, 217, 218, 219],
]

LIGHT_COLOR2: List[List[int]] = [
    [40, 41, 42, 43, 44, 45],
    [76, 77, 78, 79, 80, 81],
    [112, 113, 114, 115, 116, 117],
    [148, 149, 150, 151, 152, 153],
    [184, 185, 186, 187, 188, 189],
    [220, 221, 222, 223, 224, 225],
]

LIGHT_COLOR3: List[List[int]] = [
    [46, 47, 48, 49, 50, 51],
    [82, 83, 84, 85, 86, 87],
    [118, 119, 120, 121, 122, 123],
    [154, 155, 156, 157, 158, 159],
    [190, 191, 192, 193, 194, 195],
    [226, 227, 228, 229, 230, 231],
]

color_sets = {
    "dark": {
        "color1": DARK_COLOR1,
        "color2": DARK_COLOR2,
        "color3": DARK_COLOR3,
        "greyscale": DARK_GREYSCALE,
    },
    "light": {
        "color1": LIGHT_COLOR1,
        "color2": LIGHT_COLOR2,
        "color3": LIGHT_COLOR3,
        "greyscale": LIGHT_GREYSCALE,
    },
}


def is_dark(code: int) -> bool:
    """Returns True if the given color is a dark color, False otherwise."""
    return code in DARK_COLORS


def colorbox(code: int, background: bool = False) -> str:
    """Returns a padded, styled box for the given color. If the color is
    too dark, the text is white, otherwise dark grey.
    """

    # if color is too dark, use white text, otherwise black
    text_color = 15 if is_dark(code) else 0

    text = f"{str(code).rjust(4)} "

    if background:
        return f"\033[0m\033[48;5;{code}m\033[38;5;{text_color}m{text}\033[0m"
    else:
        return f"\033[0m\033[38;5;{code}m{text}\033[0m"


def discover(background: bool = False):
    """Prints all 256 colors in a matrix on your system."""

    print("Standard: ", end="")
    for col in STANDARD:
        print(colorbox(col, background), end="")

    print("\nBright:   ", end="")
    for col in BRIGHT:
        print(colorbox(col, background), end="")

    print("\n")

    def print_row(brightness: Literal["light", "dark"], cset: int, row: int, i: int):
        cset_key = "color" + str(cset)
        col = (row * 6) + i
        colorset = cast(List[List[int]], color_sets[brightness][cset_key])
        print(
            colorbox(
                flatmap(colorset)[col],
                background,
            ),
            end="",
        )

    for cset in range(1, 4):
        # printing 18 lines of colors:
        # [16, 17, 18, 19, 20, 21]   [34, 35, 36, 37, 38, 39]
        # [52, 53, 54, 55, 56, 57]   [70, 71, 72, 73, 74, 75]
        # .. etc. for the other 4 color sets

        for row in range(6):
            for i in range(6):
                print_row("dark", cset, row, i)
            print(str().rjust(1 if background else 4), end="")
            for i in range(6):
                print_row("light", cset, row, i)
            print("")
        print("")

    print("Greyscale: ", end="")
    for grey in DARK_GREYSCALE:
        print(colorbox(grey, background), end="")

    print("\n           ", end="")
    for grey in LIGHT_GREYSCALE:
        print(colorbox(grey, background), end="")

    print("\n")



#!/usr/bin/env python

# Tinta
# Copyright 2024 github.com/brandoncript

# This program is bound to the Hippocratic License 2.1
# Full text is available here:
# https://firstdonoharm.dev/version/2/1/license

# Further to adherence to the Hippocratic License, permission is hereby
# granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software") under the terms of the
# MIT License to deal in the Software without restriction, including without
# limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and / or sell copies of the Software, and to permit persons
# to whom the Software is furnished to do so, subject to the conditions layed
# out in the MIT License.

# Where a conflict or dispute would arise between these two licenses, HLv2.1
# shall take precedence.

from logging import getLogger

__version__ = "0.1.7b4-post0"

logger = getLogger(__name__)

try:
    from .tinta import Tinta

    assert bool(Tinta)
except ImportError as e:
    logger.warning(e)
    raise e



# pylint: disable=import-error
from tinta.discover import color_sets, discover, flatmap, is_dark


class TestDiscover:

    def test_is_dark_or_light(self):

        dark_colors = [
            *flatmap(color_sets["dark"]["color1"]),  # type: ignore
            *flatmap(color_sets["dark"]["color2"]),  # type: ignore
            *flatmap(color_sets["dark"]["color3"]),  # type: ignore
            *color_sets["dark"]["greyscale"],
        ]  # type: ignore

        light_colors = [
            *flatmap(color_sets["light"]["color1"]),  # type: ignore
            *flatmap(color_sets["light"]["color2"]),  # type: ignore
            *flatmap(color_sets["light"]["color3"]),  # type: ignore
            *color_sets["light"]["greyscale"],
        ]  # type: ignore

        for col in dark_colors:
            assert is_dark(col)

        for col in light_colors:
            assert not is_dark(col)

    def test_discover(self):

        print("Testing discover()...")
        discover(background=False)

        print("Testing discover(background=True)...")
        discover(background=True)



#!/usr/bin/env python

# Tinta
# Copyright 2024 github.com/brandoncript

# This program is bound to the Hippocratic License 2.1
# Full text is available here:
# https://firstdonoharm.dev/version/2/1/license

# Further to adherence to the Hippocratic License, permission is hereby
# granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software") under the terms of the
# MIT License to deal in the Software without restriction, including without
# limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and / or sell copies of the Software, and to permit persons
# to whom the Software is furnished to do so, subject to the conditions layed
# out in the MIT License.

# Where a conflict or dispute would arise between these two licenses, HLv2.1
# shall take precedence.

"""Begin import shim
Adds the tinta path so we can load the module directly. You won't
need to do this in your project, because the package will have
been installed via pip into the correct modules dir.
"""
import sys
from pathlib import Path
from typing import Iterable, List

sys.path.append(str(Path().cwd().parent / "tinta"))
# pylint: disable=wrong-import-position, wrong-import-order, import-error
from tinta import Tinta  # noqa: E402

# End import shim


def header(s):
    print("\n\n" + s.center(80, "-") + "\n")


def sort_colors(colors: Iterable[str]) -> List[str]:
    return list(
        sorted(
            colors,
            key=lambda x: (
                "gre" if "gray" in x or "grey" in x else x.split("_")[-1].lower()
            ),
        )
    )


def complete():

    _orig_colors_ini = Tinta.colors._colors_ini_path

    Tinta.load_colors("examples/colors.ini")

    # from colors.ini:

    # [colors]
    # green = 35
    # red = 1
    # blue = 32
    # light_blue = 37
    # yellow = 214
    # amber = 208
    # olive = 106
    # orange = 166
    # purple = 18
    # pink = 197
    # gray = 243
    # dark_gray = 235
    # light_gray = 248
    # black = 0
    # white = 255
    # error = 1
    # debug = 160

    # get names of colors from colors.ini
    colors = []
    with open("examples/colors.ini", "r", encoding="utf-8") as f:
        colors = [line.split("=")[0].strip() for line in f.readlines() if "=" in line]

    for color in colors.copy():
        if "gray" in color:
            # insert grey aliases
            colors.append(color.replace("gray", "grey"))

    colors = sort_colors(colors)

    w = 28
    header("Color methods")
    GAP = "\n\n"

    method = 'Tinta().print("plain")'.ljust(w)
    Tinta(method, " → plain").print(end=GAP)

    for col in colors:
        viz = f" → {col}"
        method = f'Tinta().tint("{col}")'.ljust(w)
        Tinta().tint(col, method, viz).print()

        t = Tinta()
        func = getattr(t, col)
        method = f't.{col}("{col}")'.ljust(w)
        func(method, viz).print(end=GAP)

    w = 32

    header("Style methods")

    method = 'Tinta().bold("bold")'.ljust(w)
    Tinta(method, " →").bold("bold").print()

    method = 'Tinta().underline("underline")'.ljust(w)
    Tinta(method, " →").underline("underline").print()

    method = 'Tinta()._("underline")'.ljust(w)
    Tinta(method, " →")._("underline").print()

    # dim
    method = 'Tinta().dim("dim")'.ljust(w)
    Tinta(method).dim(" → dim").print(end=GAP)

    header("Chaining methods (builder pattern)")

    # chain methods
    t = Tinta()
    for i, col in enumerate(colors):
        t.tint(col, col, sep="\n" if i % 2 != 0 else " ")
    t.print(end=GAP)

    header("Mixed formatting")

    # complex formatting
    t = Tinta()
    t.vanilla("vanilla").bold("bold", sep="\n")
    t.clear()
    t.mint("mint").underline("underline", sep="\n")
    t.clear()
    t.olive("olive").dim("dim", sep="\n")

    t.print(end=GAP)

    header("Clearing the console")

    # clear
    t = Tinta()
    t.vanilla("vanilla").bold("bold", sep="\n")
    t.clear("plain text", sep="\n")
    t.mint("mint").underline("underline", sep="\n")
    t.olive("olive inherits underline", sep="\n").dim("dim inherits both", sep="\n")
    t.clear("clear clears all", sep="\n")
    t.amber("so we can start fresh")

    t.print()

    header("Tint via color codes")

    def print_columns(cols, *, width=4, indent=0):
        print("".ljust(width + 1) * indent, end="")
        for i in range(256):
            end = "\n" if i % cols == cols - indent - 1 else " "
            Tinta().tint(f"{str(i).ljust(width)}", color=i).print(end=end)

    print_columns(12, indent=8)

    Tinta.colors._colors_ini_path = _orig_colors_ini


if __name__ == "__main__":
    complete()



#!/usr/bin/env python

# Tinta
# Copyright 2024 github.com/brandoncript

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# If you use this software, you must also agree under the terms of the Hippocratic License 3.0 to not use this software in a way that directly or indirectly causes harm. You can find the full text of the license at https://firstdonoharm.dev.

"""This module contains type hints and utility functions for Tinta."""
import functools
from typing import Any, Callable, cast

from .multi_version_imports import GenericCallable, Literal

StringType = Literal["pln", "esc", "fmt"]  # type: ignore


class MissingColorError(Exception):
    """Raised when a color is not found in the colors.ini file."""

    pass


def copy_kwargs(func: GenericCallable) -> Callable[..., GenericCallable]:
    """Decorator does nothing but casts the original function to match the given function signature"""

    @functools.wraps(func, updated=())
    def _cast_func(_func: Callable[..., Any]) -> GenericCallable:
        return cast(GenericCallable, _func)

    if not callable(func):
        raise RuntimeError(
            f"You must pass a function to this decorator, got {func} instead."
        )

    return _cast_func


def parse_bool(value: Any) -> bool:
    """Parses a string value to a boolean value."""
    if isinstance(value, bool):
        return value
    return str(value).lower() in ("true", "1", "t", "y", "yes")



#!/usr/bin/env python

# Tinta
# Copyright 2024 github.com/brandoncript

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# If you use this software, you must also agree under the terms of the Hippocratic License 3.0 to not use this software in a way that directly or indirectly causes harm. You can find the full text of the license at https://firstdonoharm.dev.

"""This class is a low-level helper class for managing colors in Tinta."""

import configparser
import sys
from pathlib import Path
from typing import List, Optional, Union

from .typ import MissingColorError

config = configparser.ConfigParser()


def _alias_key(colors: "AnsiColors", k: str, search: str, repl: str):
    """Sets up an alias key for a color."""
    if k not in config["colors"] and k not in colors.__dict__:
        raise MissingColorError(f"Color '{k}' not found in colors.ini.")
    if search.lower() in k.lower():
        alias_key = k.replace(search.lower(), repl.lower())
        if alias_key not in config["colors"] and alias_key not in colors.__dict__:
            colors.__setattr__(alias_key, int(config["colors"][k]))
            config["colors"][alias_key] = config["colors"][k]


class AnsiColors:
    """Color builder for Tinta's console output.

    ANSI color map for console output. Get a list of colors here =
    http://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html#256-colors

    Or run Tinta.discover() to see all 256 colors on your system.

    You can change the colors the terminal outputs by changing the
    ANSI values in colors.ini.
    """

    def __init__(self, path: Optional[Union[str, Path]] = None):
        path = Path(path) if path else Path(__file__).parent / "colors.ini"
        if not path.is_absolute():
            path = Path().cwd() / path
        if not path.exists():
            raise FileNotFoundError(
                f"Tinta failed to load colors, '{path}' does not exist."
            )

        # if path is a dir, look for colors.ini
        if path.is_dir():
            path = path / "colors.ini"

        # if there is still no colors.ini file, look for one in Path.cwd() or PYTHONPATH
        if not path.exists():
            for p in [Path().cwd(), Path(sys.path[0])]:
                if (p / "colors.ini").exists():
                    path = p / "colors.ini"
                    break

        if not path.exists():
            raise FileNotFoundError(
                f"Tinta failed to load colors, could not find 'colors.ini' in cwd or in PYTHONPATH. Please provide a valid path to a colors.ini file."
            )

        config.read(path)
        for k, v in config["colors"].items():
            self.__setattr__(k, int(v))

            _alias_key(self, k, "gray", "grey")
            _alias_key(self, k, "grey", "gray")

        self._colors_ini_path = path

    def get(self, color: str) -> int:
        """Returns the ANSI code for a color.

        Args:
            color (str): A color name.

        Returns:
            int: The ANSI code for the color.
        """

        if color == "default":
            return 0

        if color not in config["colors"]:
            raise MissingColorError(f"Color '{color}' not found in colors.ini.")

        return int(config["colors"][color])

    def reverse_get(self, code: int) -> str:
        """Returns the color name for an ANSI code.

        Args:
            code (int): An ANSI code.

        Returns:
            str: The color name for the code.
        """
        for k, v in config["colors"].items():
            if int(v) == code:
                return k

        raise MissingColorError(
            f"Color with ANSI code '{code}' not found in colors.ini."
        )

    def list_colors(self) -> List[str]:
        """Returns a list of all colors in the colors.ini file."""
        return list(config["colors"].keys())



#!/usr/bin/env python

# Tinta
# Copyright 2024 github.com/brandoncript

# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# If you use this software, you must also agree under the terms of the Hippocratic License 3.0 to not use this software in a way that directly or indirectly causes harm. You can find the full text of the license at https://firstdonoharm.dev.

"""Tinta is a magical console output tool with support for printing in beautiful
colors and with rich formatting, like bold and underline. It's so pretty,
it's almost like a unicorn.
"""

import configparser
import functools
import os
import re
import sys
from itertools import zip_longest
from pathlib import Path
from typing import Any, cast, Dict, List, Optional, overload, Union

from .ansi import AnsiColors
from .colorize import colorize, ensure_reset, tint, was_reset
from .constants import (
    CURSOR_UP_ONE,
    ERASE_LINE,
    PREFER_PLAINTEXT,
    SEP,
    SMART_FIX_PUNCTUATION,
    STEALTH,
)
from .discover import discover as _discover
from .multi_version_imports import deprecated
from .typ import copy_kwargs, MissingColorError, StringType

config = configparser.ConfigParser()


class _MetaTinta(type):

    def __init__(cls, name, bases, dct):
        super(_MetaTinta, cls).__init__(name, bases, dct)
        cls.colors = AnsiColors()

    def load_colors(cls, path: Union[str, Path]):
        loaded_colors = AnsiColors(path)

        # Check if any of the color names loaded match the built-in methods. If so, raise an error.
        for col in loaded_colors.list_colors():
            if hasattr(cls, col):
                raise AttributeError(
                    f"Cannot overwrite built-in method '\
                        {col}' with color name. Please rename the color in '{path}'."
                )

        # Add the loaded colors to the Tinta class
        cls.colors = loaded_colors


class Tinta(metaclass=_MetaTinta):
    """Tinta is a magical console output tool with support for printing in
    beautiful colors and with rich formatting, like bold and underline. It's
    so pretty, it's almost like a unicorn.

    All public methods chain together to form a builder pattern, e.g.:

    (Tinta('Some plain text')
        .white(' white')
        .blue(' blue')
        .red(' red')
        .bold().red(' bold red')
        .dark_gray()
        .dim(' dim').print())

    Args:
        *s (tuple(Any)): A sequence of one or more text strings, to be joined together.
        sep (str): Used to join segment strings. Defaults to ' '.

    Attributes:
        color (str, int):           A color string or ansi color code (int),
                                    e.g. 'white' or 42
        style (str):                A style string, e.g. 'bold', 'dim', 'underline'.
                                    Multiple styles are joined with a +
        parts (list):               A list of Part objects.
        get_parts(style: str):      Returns a list of string parts, or a list of Parts.

    Methods:
        to_str() -> str:           Returns a compiled rich text string, or use plaintext=True for plaintext.
        push() -> self:             Adds segments to this Tinta instance.
        pop(qty: int) -> self:              Removes the last 'qty' segments from this Tinta instance.
        text()                      [deprecated]
        plaintext()                 [deprecated]
        add()                       [deprecated, use push or __call__]
        code()                      [deprecated]
        bold() -> self:             Sets segments to bold.
        underline() -> self:        Sets segments to underline.
        dim() -> self:              Sets segments to a darker, dimmed color.
        normal() -> self:           Removes all styles.
        clear() -> self:            Removes all styles and colors.
        line() -> self:             Adds segments on a new line.
        print():                    Prints the output of a Tinta instance, then clears.
    """

    _initialized = False
    _known_colors: Dict[str, Any] = {}
    color: Union[int, str]
    colors: AnsiColors

    def __init__(
        self, *s: Any, color: Optional[Union[str, int]] = None, sep: str = SEP
    ):
        """Main intializer for Tinta

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.
        """

        self.color = color or 0  # 0 is the default color for terminals
        self.style: List[str] = []
        self._parts: List["Tinta.Part"] = []
        self._prefixes: List[str] = []

        # Inject ANSI helper functions
        if not Tinta._initialized:
            Tinta._known_colors = vars(self.colors)
            Tinta._initialized = True
        for c in Tinta._known_colors:
            self.__setattr__(c, functools.partial(self.tint, color=c))

        if s:
            self.push(*s, sep=sep)

    def __call__(self, *s: Any, sep: str = SEP) -> "Tinta":
        return self.push(*s, sep=sep)

    def __repr__(self) -> str:
        """Generates a string representation of the current
        Tinta instance.

        Returns:
            str: Plaintext string
        """
        return str(self.to_str(plaintext=True))

    def __str__(self) -> str:
        """Generates a string representation of the current
        Tinta instance, colorized.

        Returns:
            str: Colorized string"""

        return self.to_str()

    def _get_sep(
        self,
        p: "Tinta.Part",
        next_p: Optional["Tinta.Part"],
        sep: Optional[str],
    ) -> str:
        if not next_p:
            return ""
        return sep if sep is not None else p.sep

    def _get_str(
        self,
        attr: StringType,
        p: "Tinta.Part",
        next_p: Optional["Tinta.Part"] = None,
        sep: Optional[str] = None,
        fix_punc: bool = SMART_FIX_PUNCTUATION,
    ) -> str:

        s: str = getattr(p, attr)

        next_char_is_punc = (
            next_p.pln[0]
            in [
                ".",
                ",",
                ";",
                ":",
                "!",
                "?",
            ]
            if fix_punc and next_p and next_p.pln
            else False
        )

        punc_affects_sep = (
            next_char_is_punc
            and any(
                [
                    (len(next_p.pln) == 1),
                    (len(next_p.pln) > 1 and next_p.pln[1] == " "),
                ]
            )
            if next_p and next_p.pln
            else False
        )

        next_char_is_newline = bool(next_p and next_p.pln.startswith(os.linesep))
        last_char_is_newline = p.pln.endswith(os.linesep)
        newline_affects_punc = last_char_is_newline or next_char_is_newline

        should_ignore_sep = punc_affects_sep or newline_affects_punc

        if not should_ignore_sep:
            s = f"{s}{self._get_sep(p, next_p, sep)}"

        # TODO: Not certain we ever want to do this, it has too many false negatives
        # if fix_punc:
        #     s = re.sub(r"(\w)\s+([.,;:!?])", r"\1\2", s)

        if next_p is None and p.has_formatting:
            if attr == "fmt":
                s = ensure_reset(s)
            elif attr == "esc":
                s = esc(ensure_reset(s))

        return s

    def to_str(
        self,
        sep: Optional[str] = None,
        plaintext: bool = False,
        escape_ansi: bool = False,
        fix_punctuation: bool = SMART_FIX_PUNCTUATION,
    ) -> str:
        """Returns a compiled rich text string, joined by 'sep'. Note that any separators included with a
        part will be returned as part of the string - adding a separator here may result in a double separator.

        Args:
            sep (str, optional): Used to join strings. Defaults to the separator used when the part was added, or ' '. Setting this will override the part's separator.
            plaintext (bool, optional): If True, returns a plaintext string. Defaults to False.
            escape_ansi (bool, optional): If True, will escape ANSI codes' \\ → \\\\. Defaults to False.
            fix_punctuation (bool, optional): If True, will fix punctuation spacing. Defaults to SMART_FIX_PUNCTUATION or True.

        Returns:
            str: A rich text string.
        """

        if not self._parts:
            return ""

        attr: StringType = "pln" if plaintext else "esc" if escape_ansi else "fmt"

        return "".join(
            [
                self._get_str(attr, p, np, sep, fix_punctuation)
                for p, np in zip_longest(self._parts, self._parts[1:])
            ]
        )

    @deprecated("Use to_str() instead.")
    def get_text(self, sep: str = SEP) -> str:
        """
        Deprecated. Use to_str() instead.

        Args:
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            str: A rich text string.
        """
        return self.to_str(sep=sep)

    @property
    def parts(self) -> List["Tinta.Part"]:
        """A list of Tinta.Part objects.

        Returns:
            List[Part]: A list of Tinta.Part objects"""

        return self._parts

    @property
    def has_formatting(self) -> bool:
        """Returns True if any part has formatting.

        Returns:
            bool: True if any part has formatting."""
        return any(p.has_formatting for p in self._parts)

    @property
    @deprecated("Use parts_fmt instead.")
    def parts_formatted(self) -> list:
        """Returns a list of richly formated string parts

        Returns:
            str: A list of formatted strings."""
        return [part.fmt for part in self._parts]

    @property
    def parts_fmt(self) -> list:
        """Returns a list of richly formated string parts

        Returns:
            str: A list of formatted strings."""
        return [part.fmt for part in self._parts]

    @property
    @deprecated("Use parts_pln instead.")
    def parts_plaintext(self) -> list:
        """Returns a list of plaintext string parts

        Returns:
            str: A list of plaintext strings."""
        return [part.pln for part in self._parts]

    @property
    def parts_pln(self) -> list:
        """Returns a list of plaintext string parts

        Returns:
            str: A list of plaintext strings."""
        return [part.pln for part in self._parts]

    @property
    def parts_esc(self) -> list:
        """Returns a list of escaped formatted string parts, so that special characters are visible (i.e., backslashes are doubled).

        Returns:
            str: A list of esc strings."""
        return [part.esc for part in self._parts]

    @deprecated("Use to_str(plaintext=True) instead.")
    def get_plaintext(self, sep: Optional[str] = None) -> str:
        """Deprecated. Use to_str(plaintext=True) instead.

        Args:
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            str: A plaintext string.
        """
        return self.to_str(sep=sep, plaintext=True)

    def push(self, *s: Any, sep: Optional[str] = SEP) -> "Tinta":
        """Adds segments to this Tinta instance

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """

        # If an empty set of segments is passed, skip to the
        # next segment (to prevent duplicating whitespace).
        if not s:
            return self

        if sep is None:
            sep = SEP

        # Join all s parts with the specified separator
        pln = sep.join([str(x) for x in s])

        # Collect any prefixes that may have been set
        if self._prefixes:
            pln = "".join(self._prefixes) + pln
            self._prefixes = []

        # Generate style string
        style = "+".join(list(set(self.style))) if self.style else None

        fg: Union[int, None] = 0
        if isinstance(self.color, int):
            fg = self.color
        elif isinstance(self.color, str):
            fg = self.colors.get(self.color or "default")

        if not self.has_formatting and fg == 0:
            # We don't want to add an ansi reset code if we're not adding any formatting
            fg = None

        # TODO: Add support for background colors

        # Generate formatted string
        fmt = colorize(pln, fg=fg, style=style)

        # Append to parts list
        self._parts.append(Tinta.Part(fmt, pln, sep))

        return self

    @deprecated(
        "Use use push() or __call__() directly (e.g. Tinta('text')('more text')) instead."
    )
    def add(self, *s: Any, sep: str = SEP) -> "Tinta":
        """Deprecated. Use push() or just __call__() (e.g. Tinta('text')('more text')) instead.

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        return self.push(*s, sep=sep)

    def pop(self, qty: int = 1) -> "Tinta":
        """Removes the last 'qty' segments from this Tinta instance

        Args:
            qty (int, optional): Number of segments to remove. Defaults to 1.

        Returns:
            self
        """
        for _ in range(min(qty, len(self._parts))):
            self._parts.pop()
        return self

    @deprecated("Use pop() instead.")
    def remove(self, qty: int = 1) -> "Tinta":
        return self.pop(qty)

    @overload
    def tint(self, *s: Any, color: Union[str, int], sep: str = SEP) -> "Tinta": ...

    @overload
    def tint(self, color: Union[str, int], *s: Any, sep: str = SEP) -> "Tinta": ...

    # pylint: disable=redefined-outer-name
    def tint(self, *args, **kwargs) -> "Tinta":
        """Adds segments of text colored with the specified color.
        Can be used in place of calling named color methods.

        Args:
            color (str | int, optional): A color name or ANSI color index. Defaults to first argument.
            *s: Any: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """

        return tint(self, *args, **kwargs)

    @overload
    def inspect(self, code: int, name: None = None, throw: bool = False) -> str: ...

    @overload
    def inspect(
        self, code: None = None, name: str = "default", throw: bool = False
    ) -> int: ...

    def inspect(
        self,
        code: Optional[int] = None,
        name: Optional[str] = None,
        throw: bool = False,
    ) -> Union[str, int, None]:
        """Queries the known colors by code or name, and returns the corresponding name or code, respectively.

        Args:
            code (int, optional): An ANSI color code. Defaults to None.
            name (str, optional): A color name. Defaults to None.

        Returns:
            Union[str, int]: A color name or code, if found, otherwise None.
        """

        if code is not None and name is not None:
            raise AttributeError(
                "Tinta.inspect() accepts either a code or a name, not both."
            )

        try:
            if code is not None:
                if code == 0:
                    return "default"
                return self.colors.reverse_get(code)

            elif name is not None:
                if name == "default":
                    return 0
                return self.colors.get(name)
        except MissingColorError as e:
            if throw:
                raise e
        return None

    @deprecated("Use tint() instead, setting color=<int:A valid ANSI color code>.")
    def code(self, code: int = 0, *s: Any, sep: str = SEP) -> "Tinta":
        """Adds segments of text colored with the specified ANSI code.

        Args:
            *s: Segments of text to add.
            code (int, optional): An ANSI code. Defaults to 0.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        self.color = int(code)
        self.push(*s, sep=sep)
        return self

    def bold(self, *s: Any, sep: str = SEP) -> "Tinta":
        """Adds bold segments to this Tinta instance

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        self.style.append("bold")
        self.push(*s, sep=sep)
        return self

    @copy_kwargs(bold)
    def b(self, *args, **kwargs):
        return self.bold(*args, **kwargs)

    def underline(self, *s: Any, sep: str = SEP) -> "Tinta":
        """Adds underline segments to this Tinta instance

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        self.style.append("underline")
        self.push(*s, sep=sep)
        return self

    @copy_kwargs(underline)
    def u(self, *args, **kwargs):
        return self.underline(*args, **kwargs)

    @copy_kwargs(underline)
    def _(self, *args, **kwargs):
        return self.underline(*args, **kwargs)

    def dim(self, *s: Any, sep: str = SEP) -> "Tinta":
        """Adds darker (dimmed) segments to this Tinta instance

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        self.style.append("faint")
        self.push(*s, sep=sep)
        return self

    def normal(self, *s: Any, sep: str = SEP) -> "Tinta":
        """Removes all styles, then adds segments to this Tinta instance

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        self.style = []
        # self._prefixes.append('\033[24m\033[21m')
        self.push(*s, sep=sep)
        return self

    def clear(self, *s: Any, sep: str = SEP) -> "Tinta":
        """Removes all styles and colors, then adds segments to this
        Tinta instance

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        self.color = 0
        self.normal(*s, sep=sep)
        return self

    @deprecated("Use clear() instead.")
    def reset(self, *s: Any, sep: str = SEP) -> "Tinta":
        return self.push(*s, sep=sep)

    @deprecated("Use nl() instead.")
    def line(self, *s, sep: str = SEP) -> "Tinta":
        return self.nl(*s, sep=sep)

    def nl(self, *s, sep: str = SEP) -> "Tinta":
        """Adds segments to this Tinta instance, preceded by a new line.

        Args:
            *s: Segments of text to add.
            sep (str, optional): Used to join strings. Defaults to ' '.

        Returns:
            self
        """
        self._prefixes = [os.linesep]
        self.push(*s, sep=sep)
        return self

    def print(
        self,
        sep: Optional[str] = None,
        end="\n",
        file=sys.stdout,
        flush=False,
        plaintext=False,
        force=False,
    ):
        """Prints a Tinta composite to the console. Once printed,
        this Tinta instance is cleared of all configuration, but can
        can continue to be used to print.

        Env: These environment variables, when set, affect Tinta globally.
            TINTA_STEALTH (not None): Hides all console output. Can be
                                      overridden by 'force'.
            TINTA_PLAINTEXT (not None): Prints all output in plaintext.

        Args:
            sep (str, optional): Used to join strings. Defaults to ' ', or whatever was used when the part was added. Setting this will override the part's separator.
            end (str, optional): String terminator. Defaults to '\n'.
            file (optional): File to write to. Defaults to sys.stdout.
            flush (bool, optional): Clears the current console line.
            plaintext (bool, optional): Prints in plaintext. Defaults to False.
            force (bool, option): Forces printing, overriding TINTA_STEALTH.
        """
        # We don't print if the TINTA_STEALTH env is set
        if STEALTH and not force:
            return

        use_plaintext = True if plaintext or PREFER_PLAINTEXT else False
        s = self.to_str(sep=sep, plaintext=use_plaintext)
        # if s.strip() does not end with reset, add it after the last non-whitespace character
        if not use_plaintext and not was_reset(s):
            s = ensure_reset(s)
        print(
            self.to_str(sep=sep, plaintext=use_plaintext),
            end=end,
            file=file,
            flush=flush,
        )

        self.clear()
        self._parts = []

    def __getattr__(self, name: str) -> "Tinta":
        """Returns a tinted segment of text.

        Args:
            name (str): Name of the color.

        Returns:
            Tinta: A Tinta instance.
        """

        if not name.startswith("_"):
            if hasattr(self.colors, name):
                return cast("Tinta", self.tint(color=name))

            else:
                try:
                    return self.__getattribute__(name)  # type: ignore
                except AttributeError as e:
                    known_colors = "\n - ".join(self.colors.list_colors())
                    known_colors = f" - {known_colors}"
                    raise AttributeError(
                        f"Attribute '{name}' not found. Did you try and access a color that doesn't exist? Available colors:\n{known_colors}"
                    ) from e

        return self.__getattribute__(name)  # type: ignore

    @staticmethod
    def discover(background=False):
        """Prints all 256 colors in a matrix on your system. If background is True,
        it will print background colors with numbers on top."""
        _discover(background)

    @staticmethod
    def clearline():
        """Clears the current printed line."""

        if sys.stdout:
            sys.stdout.write(CURSOR_UP_ONE)
            sys.stdout.write(ERASE_LINE)
            sys.stdout.flush()

    @staticmethod
    def up():
        """Moves up to the previous line."""

        if sys.stdout:
            sys.stdout.write(CURSOR_UP_ONE)
            sys.stdout.flush()

    class Part:
        """A segment part of text, intended to be joined together
        for a print statement.

        Args:
            fmt (str):      A formatted string of text.
            p (str):        A plaintext representation of fmt.
            sep (str):      Used to join segment strings. Defaults to ' '.
        """

        def __init__(self, fmt: str, pln: str, sep: str = SEP):
            self.fmt = fmt
            self.pln = pln
            self.sep = sep

        def __str__(self):
            return self.fmt

        def __repr__(self):
            return self.__str__()

        def __eq__(self, other) -> bool:
            if not isinstance(other, Tinta.Part):
                return False
            return self.fmt == other.fmt and self.pln == other.pln

        @property
        def esc(self):
            return esc(self.fmt)

        @property
        def has_formatting(self) -> bool:
            return self.fmt != self.pln

    @staticmethod
    def strip_ansi(s: str) -> str:
        """A utility method that strips ANSI escape codes from a string, converting a styled string into plaintext."""
        return re.sub(
            r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])", "", s, re.I | re.M | re.U
        )

    @classmethod
    def ljust(cls, s: str, width: int, fillchar: str = " ") -> str:
        """Returns a string left justified in a field of a specified width, accounting for ansi formatting."""
        chars_to_add = width - len(cls.strip_ansi(s))
        return f"{s}{str(fillchar or '') * chars_to_add}"

    @classmethod
    def rjust(cls, s: str, width: int, fillchar: str = " ") -> str:
        """Returns a string right justified in a field of a specified width, accounting for ansi formatting."""
        chars_to_add = width - len(cls.strip_ansi(s))
        return f"{str(fillchar or '') * chars_to_add}{s}"


def esc(string: str, replace: bool = False) -> str:
    """Returns the raw representation of a string. If replace is true,
    replace a double backslash with a single backslash."""
    r = repr(string)[1:-1]  # Strip the quotes from representation
    if replace:
        r = r.replace("\\\\", "\\")
    return r



from functools import reduce
from typing import cast, List, Sequence

from .multi_version_imports import TypeVar

T = TypeVar("T")


def flatmap(lst: Sequence[List[T]]) -> List[T]:
    """Flattens a list of lists into a single list. If the list is already flat,
    it is returned as is.
    """

    # if lst is not a nested list, return it as is
    if not any(isinstance(i, list) for i in lst):
        return cast(List[T], lst)

    return reduce(list.__add__, lst)



#!/usr/bin/env python

# Tinta
# Copyright 2024 github.com/brandoncript

# This program is bound to the Hippocratic License 2.1
# Full text is available here:
# https://firstdonoharm.dev/version/2/1/license

# Further to adherence to the Hippocratic License, permission is hereby
# granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software") under the terms of the
# MIT License to deal in the Software without restriction, including without
# limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and / or sell copies of the Software, and to permit persons
# to whom the Software is furnished to do so, subject to the conditions layed
# out in the MIT License.

# Where a conflict or dispute would arise between these two licenses, HLv2.1
# shall take precedence.

"""Begin import shim
Adds the tinta path so we can load the module directly. You won't
need to do this in your project, because the package will have
been installed via pip into the correct modules dir.
"""
import sys
import time
from pathlib import Path

sys.path.append(str(Path().cwd().parent / "tinta"))


def basic():
    # pylint: disable=wrong-import-position, wrong-import-order, import-error
    from tinta import Tinta  # noqa: E402

    # End import shim

    Tinta.load_colors("examples/colors.ini")

    # The most basic example we can get.
    Tinta("That's a really nice car!").print()

    # Prints the entire string in red.
    Tinta().red("That's a really nice red car!").print()

    # Prints the first half in blue, then the rest in red.
    Tinta().blue("That's a cool blue car").red("but not as cool as my red one").print()

    # Prints the first few words in green, separated by _*_,
    # then the final word in purple.
    Tinta().green(
        "Sometimes", "We", "Want", "To", "Join", "Words", "Differently", sep=" * "
    ).purple("Neat!").print()

    # Here we underline a word.
    Tinta().gray("It's").underline("really").normal(
        "important to be able to style things."
    ).print()

    # Here we tried to print in pink, but used the plaintext arg in print.
    # You'll notice we still support Python's multiline \ feature.
    Tinta().pink(
        "But it's equally important to be able " "to print things in plaintext, too"
    ).print(plaintext=True)

    # Let's try some f-strings.
    animal = "Tiger"
    Tinta().orange(f"Hey, we support f-strings, too! Raaarrr, said Ms. {animal}.")

    # And dimming some text.
    Tinta().blue("We can").dim("dim").normal("things").print()

    # Things getting out of hand? You can break them up easily in multiple
    # lines, without having to fiddle with \.
    tint = Tinta()
    tint.push("Sometimes we need to")
    tint.pink("break up long lines of text")
    tint.gray("to make them easier to read.")
    tint.nl("We can even write to a new line!")
    tint.print()

    # You could do the same using multiple segments, or ()
    Tinta().vanilla(
        "I like ice cream", "it comes in all sorts" "of great and yummy flavors."
    ).print()

    (
        Tinta()
        .vanilla("I like ice cream")
        .red("especially with cherries on top.")
        .print()
    )

    # When you're done printing, Tinta resets itself, but you can still
    # reuse the original variable.
    tint.push("After a print, Tinta resets itself").green()
    tint.nl("but you can still use the same initialized version.")
    tint.print()

    # Using native print()'s built-in end, we can terminate a string
    # without a newline.
    Tinta("And of course as always,").print(end="")
    Tinta(" you can print with end=''").print()

    # Not enough colors in config.yaml? Add your own on the fly!
    Tinta(
        "Did you know, you can", "write with ansi codes directly, too?", color=127
    ).print()

    # Get the string representation of the Tinta object with to_str().
    Tinta("Sometimes you just want to get the string").to_str()
    Tinta().purple("Which").pink("is").green("pretty").blue("cool").to_str()
    Tinta().vanilla("If you need it in plaintext, you can do that, too.").to_str(
        plaintext=True
    )
    t = Tinta().mint("⭐like this⭐")
    print(f"You can also use a Tinta object in an f-string directly, {t}")

    # Have some fun with separators.
    Tinta("A bird", "I like birds", sep="; ").push(
        "And also cats", "and dogs", sep=" "
    ).print(sep="\n")

    # You could get really fancy and inject some formatted text in the middle,
    # using f-strings.
    (
        Tinta()
        .mint("Fate.")
        .dark_gray("It protects")
        .underline()
        .blue(f"fools{Tinta().normal().dark_gray(',').to_str()}", sep="")
        .normal()
        .pink("little children,")
        .dark_gray("and ships named")
        .purple("Enterprise.")
        .print()
    )

    # Tinta is also smart about how we join things together. If you join
    # several objects together, it collapses repeated whitespace. You
    # can also use 'sep' to force sections to collapse.
    t = (
        Tinta()
        .pink("A section")
        .push()
        .white()
        .blue("of text", sep="")
        .green(",")
        .push()
        .purple("separated.")
    )
    t.print()

    # And finally, you can use some helper tools to clear the current
    # console and move up a line.
    Tinta().yellow("Loading...").print()
    time.sleep(1)
    Tinta.clearline()
    Tinta().green("Done").print()
    time.sleep(1)
    Tinta.up()
    Tinta().green("Done :)").print()


if __name__ == "__main__":
    basic()



# Originally published as AnsiColors
# Copyright (c) 2012 Giorgos Verigakis <verigak@gmail.com>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# This module is a modified version of the original AnsiColors module, and further
# modified by github.com/brandoncript to be used in the Tinta project. It is
# distributed under the same license as the original module, as well as the
# MIT License.

import re
from typing import Any, List, Optional, overload, Tuple, TYPE_CHECKING, Union

from .constants import ANSI_RESET_HEX, ANSI_RESET_OCT, SEP
from .typ import MissingColorError

if TYPE_CHECKING:
    from .tinta import Tinta

STYLES = (
    "none",
    "bold",
    "faint",
    "italic",
    "underline",
    "blink",
    "blink2",
    "negative",
    "concealed",
    "crossed",
)


def was_reset(s: str) -> bool:
    return s.strip().endswith(ANSI_RESET_HEX) or s.endswith(ANSI_RESET_OCT)


def ensure_reset(s: str) -> str:
    if was_reset(s):
        return s
    last_char_idx = len(s.rstrip())
    return f"{s[:last_char_idx]}{ANSI_RESET_HEX}{s[last_char_idx:]}"


def _parse_rgb(s: str) -> Tuple[int, ...]:
    if not isinstance(s, str):
        raise ValueError(f"Could not parse color '{s}'")
    s = s.strip().replace(" ", "").lower()

    # 6-digit hex
    match = re.match("#([a-f0-9]{6})$", s)
    if match:
        core = match.group(1)
        return tuple(int(core[i : i + 2], 16) for i in range(0, 6, 2))

    # 3-digit hex
    match = re.match("#([a-f0-9]{3})$", s)
    if match:
        return tuple(int(c * 2, 16) for c in match.group(1))

    # rgb(x,y,z)
    match = re.match(r"rgb\((\d+,\d+,\d+)\)", s)
    if match:
        return tuple(int(v) for v in match.group(1).split(","))

    raise ValueError(f"Could not parse color '{s}'")


def _join(*values: Union[int, str]) -> str:
    """
    Join a series of values with semicolons. The values
    are either integers or strings, so stringify each for
    good measure. Worth breaking out as its own function
    because semicolon-joined lists are core to ANSI coding.
    """
    return ";".join(str(v) for v in values)


def _color_code(
    spec: Union[str, int, Tuple[int, int, int], List[int]], base: int
) -> str:
    """
    Workhorse of encoding a color. Give preference to named colors from
    ANSI, then to specific numeric or tuple specs. If those don't work,
    try looking up look CSS color names or parsing CSS hex and rgb color
    specifications.

    :param spec: Unparsed color specification
    :param base: Either 30 or 40, signifying the base value
        for color encoding (foreground and background respectively).
        Low values are added directly to the base. Higher values use `
        base + 8` (i.e. 38 or 48) then extended codes.
    :returns: Discovered ANSI color encoding.
    :raises: ValueError if cannot parse the color spec.
    """
    if isinstance(spec, str):
        spec = spec.strip().lower()

    if spec == "default" or spec == 0:
        return "0"
    elif isinstance(spec, int) and 1 <= spec <= 255:
        return _join(base + 8, 5, spec)
    elif isinstance(spec, (tuple, list)):
        return _join(base + 8, 2, _join(*spec))
    elif str(spec).startswith("#"):
        rgb = _parse_rgb(str(spec))
        # parse_rgb raises ValueError if cannot parse spec
        return _join(base + 8, 2, _join(*rgb))
    else:
        raise ValueError(f"Could not parse color '{spec}'")


def colorize(
    s: str,
    fg: Optional[int] = None,
    bg: Optional[int] = None,
    style: Optional[str] = None,
    reset: bool = False,
) -> str:
    """
    Add ANSI colors and styles to a string.

    Args:
    s: String to format.
    fg: Foreground color specification.
    bg: Background color specification.
    style: Style names, separated by '+'
    reset: Whether to reset the formatting at the end of the string. Default is False.

    Returns: Formatted string.
    """
    codes: List[str] = []

    if fg is not None:
        codes.append(_color_code(fg, 30))
    # TODO: Background is not implemented
    # if bg is not None:
    #     codes.append(_color_code(bg, 40))
    if style is not None:
        for style_part in style.split("+"):
            if style_part in STYLES:
                codes.append(str(STYLES.index(style_part)))
            else:
                raise ValueError(f'Invalid style "{style_part}"')

    if codes:
        template = "\x1b[{0}m{1}"
        if reset:
            template += ANSI_RESET_HEX
        return template.format(_join(*codes), s)
    else:
        return s


@overload
def tint(
    instance: "Tinta", *s: Any, color: Union[str, int], sep: str = SEP
) -> "Tinta": ...


@overload
def tint(
    instance: "Tinta", color: Union[str, int], *s: Any, sep: str = SEP
) -> "Tinta": ...


# pylint: disable=redefined-outer-name
def tint(instance: "Tinta", *args, **kwargs) -> "Tinta":
    """Adds segments of text colored with the specified color.
    Can be used in place of calling named color methods.

    Args:
        instance (Tinta): The Tinta instance.
        color (str | int, optional): A color name or ANSI color index. Defaults to first argument.
        *s: Any: Segments of text to add.
        sep (str, optional): Used to join strings. Defaults to ' '.

    Returns:
        self
    """

    # color: Optional[Union[str, int]] = None, *s: Any, sep: str = SEP

    # check if the first argument is a known color or valid ANSI code, or comes from kwargs
    self = instance
    s = args
    color = kwargs.get("color", None)
    sep = kwargs.get("sep", SEP)
    if color is None:
        if not len(s) > 1:
            raise AttributeError(
                "If no color is specified, tint() requires at least two arguments."
            )
        if args and isinstance(args[0], (str, int)):
            color = s[0]
            s = s[1:]
        else:
            raise AttributeError(
                "Could not determine color from arguments. Either pass a color as the first argument, or use the color keyword argument."
            )

    # if color is numeric integer string, assume it's an ANSI color code
    if isinstance(color, int) or (isinstance(color, str) and color.isdigit()):
        self.color = int(color)

    # Check if color_name is a valid color if color is a string
    if isinstance(color, str):
        if not hasattr(self.colors, color):  # type: ignore
            raise MissingColorError(
                f"Invalid color name: {color}. Is it in colors.ini?"
            )
        self.color = self.colors.get(color)  # type: ignore

    self.push(*s, sep=sep)
    return self



#!/usr/bin/env python

# Tinta
# Copyright 2024 github.com/brandoncript

# This program is bound to the Hippocratic License 2.1
# Full text is available here:
# https://firstdonoharm.dev/version/2/1/license

# Further to adherence to the Hippocratic License, permission is hereby
# granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software") under the terms of the
# MIT License to deal in the Software without restriction, including without
# limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and / or sell copies of the Software, and to permit persons
# to whom the Software is furnished to do so, subject to the conditions layed
# out in the MIT License.

# Where a conflict or dispute would arise between these two licenses, HLv2.1
# shall take precedence.


import os
import re
from contextlib import contextmanager
from typing import Any, Callable, Dict, List, Tuple

import pytest
from pytest import CaptureFixture

# pylint: disable=import-error
from tinta import Tinta

Tinta.load_colors("examples/colors.ini")

O = "\x1b[0m"
GREEN = "\x1b[38;5;35m"
RED = "\x1b[38;5;1m"
BLUE = "\x1b[38;5;32m"
L_BLUE = "\x1b[38;5;37m"
YELLOW = "\x1b[38;5;214m"
AMBER = "\x1b[38;5;208m"
MINT = "\x1b[38;5;43m"
OLIVE = "\x1b[38;5;106m"
ORANGE = "\x1b[38;5;166m"
PURPLE = "\x1b[38;5;18m"
PINK = "\x1b[38;5;197m"
GRAY = "\x1b[38;5;243m"
D_GRAY = "\x1b[38;5;235m"
L_GRAY = "\x1b[38;5;248m"
WHITE = "\x1b[38;5;255m"


@contextmanager
def skip_on_github_actions():
    gh = bool(os.getenv("GITHUB_ACTIONS") == "true")
    if gh:
        pytest.mark.skipif(True, reason="Function skipped on GitHub Actions")
    yield gh


class TestInit:

    def test_init(self):
        assert len(Tinta("initialized").parts) == 1


class TestBasicColorizing:

    @pytest.mark.parametrize(
        "color,Testa,expected",
        [
            # fmt: off
            ("green", lambda: Tinta().green("green"), f"{GREEN}green{O}"),
            ("green", lambda: Tinta().tint(35, "green"), f"{GREEN}green{O}"),
            ("green", lambda: Tinta().tint("green", "green"), f"{GREEN}green{O}"),
            ("red", lambda: Tinta().red("red"), f"{RED}red{O}"),
            ("blue", lambda: Tinta().blue("blue"), f"{BLUE}blue{O}"),
            ("blue", lambda: Tinta().blue("Green"), f"{BLUE}Green{O}"),
            ("light_blue", lambda: Tinta().light_blue("light_blue"), f"{L_BLUE}light_blue{O}"),
            ("yellow", lambda: Tinta().yellow("yellow"), f"{YELLOW}yellow{O}"),
            ("amber", lambda: Tinta().amber("amber"), f"{AMBER}amber{O}"),
            ("mint", lambda: Tinta().mint("Mint"), f"{MINT}Mint{O}"),
            ("olive", lambda: Tinta().olive("olive"), f"{OLIVE}olive{O}"),
            ("orange", lambda: Tinta().orange("orange"), f"{ORANGE}orange{O}"),
            ("purple", lambda: Tinta().purple("purple"), f"{PURPLE}purple{O}"),
            ("pink", lambda: Tinta().pink("pink"), f"{PINK}pink{O}"),
            ("gray", lambda: Tinta().gray("gray"), f"{GRAY}gray{O}"),
            ("dark_gray", lambda: Tinta().dark_gray("dark_gray"), f"{D_GRAY}dark_gray{O}"),
            ("light_gray", lambda: Tinta().light_gray("light_gray"), f"{L_GRAY}light_gray{O}"),
            ("white", lambda: Tinta().white("white"), f"{WHITE}white{O}"),
            # fmt: on
        ],
    )
    def test_colors(self, color, Testa, expected):
        Testa().print()
        assert Testa().to_str() == expected


class TestChaining:

    @pytest.mark.parametrize(
        "Testa, kwargs, expected",
        [
            # fmt: off
            (lambda: Tinta().green("green").red("red").blue("blue"), {"sep":"", "plaintext":True}, "greenredblue"),
            (lambda: Tinta().green("green").red("red").blue("blue"), {"plaintext": True}, "green red blue"),
            (lambda: Tinta().green("green").red("red").blue("blue"), {"sep": ""}, f"{GREEN}green{RED}red{BLUE}blue{O}"),
            (lambda: Tinta().green("green").red("red").blue("blue"), {}, f"{GREEN}green {RED}red {BLUE}blue{O}"),
            (lambda: Tinta().green("green").red("red").blue("blue "), {}, f"{GREEN}green {RED}red {BLUE}blue{O} "),
            (lambda: Tinta().green("green").red("red").blue("blue \n"), {}, f"{GREEN}green {RED}red {BLUE}blue{O} \n"),
            (lambda: Tinta().green("green").red("red").blue("blue").clear(), {}, f"{GREEN}green {RED}red {BLUE}blue{O}"),
            # fmt: on
        ],
    )
    def test_chaining_resets_correctly(
        self,
        Testa: Callable,
        kwargs: Dict[str, Any],
        expected: str,
        capfd: CaptureFixture[str],
    ):
        s = Testa().to_str(**kwargs)
        Testa().print(**kwargs)
        assert s == expected
        with skip_on_github_actions() as skip:
            if not skip:
                out = capfd.readouterr().out
                assert out == f"{expected}\n"

    @pytest.mark.parametrize(
        "Testa,expected",
        [
            (
                lambda: Tinta().mint("Mint\nice cream"),
                f"{MINT}Mint\nice cream{O}",
            ),
            (
                lambda: Tinta().mint("Mint").white("\nice cream\nis the")._("\nbest"),
                f"{MINT}Mint{WHITE}\nice cream\nis the\x1b[38;5;255;4m\nbest{O}",
            ),
            (
                lambda: Tinta("\n").mint("Mint\n").white("ice cream is the")._("best"),
                f"\n{MINT}Mint\n{WHITE}ice cream is the \x1b[38;5;255;4mbest{O}",
            ),
        ],
    )
    def test_newlines_in_chaining(self, Testa: Callable[[], Tinta], expected: str):
        s = Testa().to_str()
        Testa().print()
        assert s == expected


class TestUnicode:

    def test_unicode(self):
        Tinta().pink("I ♡ Unicorns").print()
        assert Tinta().pink("I ♡ Unicorns").to_str() == f"{PINK}I ♡ Unicorns{O}"

    def test_box_drawing(self):
        s = f"┌{'─'*14}┐\n│ I ♡ Unicorns │\n└{'─'*14}┘"
        Tinta("\n").mint(s).print()
        assert (
            Tinta().mint(s).to_str()
            == f"{MINT}┌──────────────┐\n│ I ♡ Unicorns │\n└──────────────┘{O}"
        )

    def test_emoji(self):
        Tinta("🦄").print()
        assert Tinta("🦄").to_str() == "🦄"

    def test_emoji_with_color(self):
        Tinta().purple("🦄").print()
        assert Tinta().purple("🦄").to_str() == f"{PURPLE}🦄{O}"

    def test_emoji_with_color_and_clear(self):
        Tinta().pink("🦄").clear("🦄").print()
        assert Tinta().pink("🦄").clear("🦄").to_str(sep="") == f"{PINK}🦄\x1b[0m🦄{O}"

    def test_emoji_with_color_and_clear_and_emoji(self):
        Tinta().purple("🦄").clear("🦄").pink("🦄").print()
        assert (
            Tinta().purple("🦄").clear("🦄").pink("🦄").to_str()
            == f"{PURPLE}🦄 \x1b[0m🦄 {PINK}🦄{O}"
        )


class TestLowerLevel:

    ITS_NOT_EASY = "it's not easy being green"

    def assert_its_not_easy_being_green(self, out: str):
        assert self.ITS_NOT_EASY in out
        assert f"{GREEN}" in out

    def test_tint_takes_int_arg0_as_color(self):
        out = Tinta().tint(35, self.ITS_NOT_EASY).to_str()
        self.assert_its_not_easy_being_green(out)

    def test_tint_takes_str_arg0_as_color(self):
        out = Tinta().tint("green", self.ITS_NOT_EASY).to_str()
        self.assert_its_not_easy_being_green(out)

    def test_tint_takes_int_color_kwarg(self):
        out = Tinta().tint(self.ITS_NOT_EASY, color=35).to_str()
        self.assert_its_not_easy_being_green(out)

    def test_tint_takes_str_color_kwarg(self):
        out = Tinta().tint(self.ITS_NOT_EASY, color="green").to_str()
        self.assert_its_not_easy_being_green(out)


class TestEdgeCases:

    def test_missing_color(self):
        with pytest.raises(Exception):
            Tinta().sparkle().print()

    def test_color_same_as_builtin_method(self):
        with pytest.raises(Exception):
            Tinta().load_colors("tests/test_colors_invalid.ini")

    def test_print_empty(self):
        Tinta().print()

    def test_print_empty_str(self):
        Tinta("").print()

    def test_print_none(self):
        Tinta(None).print()

    def test_empty_color_call(self):
        t = Tinta("Plain").green()
        t.push("Green")
        t.print()

    def test_tint_color_0(self):
        Tinta().tint(0, "Zero").print()


class TestExamples:

    def test_basic_example(self):
        from examples.basic_example import basic

        basic()

    def test_complete_example(self):
        from examples.complete_example import complete

        complete()


class TestWhiteSpace:

    def test_print_whitespace(self):
        Tinta(" ").print()

    whitespace_cases = [
        (lambda: Tinta("Hello\nWorld"), "Hello\nWorld"),
        (lambda: Tinta("\nHello\nWorld"), "\nHello\nWorld"),
        (lambda: Tinta("\nHello\nWorld\n"), "\nHello\nWorld\n"),
        (lambda: Tinta("\nHello\nWorld\n\n"), "\nHello\nWorld\n\n"),
        (lambda: Tinta("Hello\nWorld\n"), "Hello\nWorld\n"),
        (lambda: Tinta("Hello World\n"), "Hello World\n"),
        (lambda: Tinta("Hello World"), "Hello World"),
        (lambda: Tinta("Hello World "), "Hello World "),
        (lambda: Tinta(" Hello World "), " Hello World "),
        (lambda: Tinta(" Hello World"), " Hello World"),
        (lambda: Tinta("  Hello World"), "  Hello World"),
        (lambda: Tinta("  Hello World "), "  Hello World "),
        (lambda: Tinta("  Hello World  "), "  Hello World  "),
        (lambda: Tinta(" Hello World\n"), " Hello World\n"),
        (lambda: Tinta(" Hello World\n "), " Hello World\n "),
        (lambda: Tinta(" Hello World\n  "), " Hello World\n  "),
        (lambda: Tinta(" Hello World\n\n"), " Hello World\n\n"),
        (lambda: Tinta(" Hello World\n\n "), " Hello World\n\n "),
        (lambda: Tinta("\nHello World \n"), "\nHello World \n"),
        (lambda: Tinta("\nHello World \n "), "\nHello World \n "),
        (lambda: Tinta("\n Hello World \n\n"), "\n Hello World \n\n"),
        (lambda: Tinta("\n Hello\n World \n"), "\n Hello\n World \n"),
    ]

    @pytest.mark.parametrize(
        "Testa,expected",
        whitespace_cases,
    )
    def test_uncolored_respects_whitespace(
        self, Testa: Callable[[], Tinta], expected: str, capfd: CaptureFixture[str]
    ):
        s = Testa().to_str()
        assert s == expected
        Testa().print()
        with skip_on_github_actions() as skip:
            if not skip:
                out = capfd.readouterr().out
                assert out == f"{expected}\n"

    @pytest.mark.parametrize(
        "Testa,expected",
        whitespace_cases,
    )
    def test_uncolored_plaintext_respects_whitespace(
        self, Testa: Callable[[], Tinta], expected: str, capfd: CaptureFixture[str]
    ):
        p = Testa().to_str(plaintext=True)
        assert p == expected
        Testa().print(plaintext=True)
        with skip_on_github_actions() as skip:
            if not skip:
                out = capfd.readouterr().out
                assert out == f"{expected}\n"

    @pytest.mark.parametrize(
        "Testa,expected",
        whitespace_cases,
    )
    def test_proxy_color_to_str_respects_whitespace(
        self, Testa: Callable[[], Tinta], expected: str, capfd: CaptureFixture[str]
    ):
        s = Testa().to_str()
        found = re.search(r"\s+$", expected)
        trail_ws = found.group() if found else ""
        _expected = f"{PURPLE}{expected.rstrip()}{O}{trail_ws}"
        #                                         ^^^^^^^^^^^^^
        # Reset ANSI char is always placed before trailing whitespace
        assert Tinta().purple(s).to_str() == _expected
        Tinta().purple(s).print()
        with skip_on_github_actions() as skip:
            if not skip:
                out = capfd.readouterr().out
                assert out == f"{_expected}\n"

    @pytest.mark.parametrize(
        "Testa,expected",
        whitespace_cases,
    )
    def test_proxy_color_to_plaintext_respects_whitespace(
        self, Testa: Callable[[], Tinta], expected: str, capfd: CaptureFixture[str]
    ):
        p = Testa().to_str(plaintext=True)
        assert Tinta().purple(p).to_str(plaintext=True) == expected
        Tinta().purple(p).print(plaintext=True)
        with skip_on_github_actions() as skip:
            if not skip:
                out = capfd.readouterr().out
                assert out == f"{expected}\n"

    @pytest.mark.parametrize(
        "Testa,expected",
        whitespace_cases,
    )
    def test_proxy_color_chain_to_str_respects_whitespace(
        self, Testa: Callable[[], Tinta], expected: str, capfd: CaptureFixture[str]
    ):
        s = Testa().to_str()
        found = re.search(r"\s+$", expected)
        trail_ws = found.group() if found else ""
        expected_purple = f"{PURPLE}{expected}"
        expected_pink = f"{PINK}{expected.rstrip()}{O}{trail_ws}"
        l_or_r_newline = expected.startswith("\n") or expected.endswith("\n")
        sep = "" if l_or_r_newline else " "
        # ^ If there is a newline at the end of the current or start of the next segment,
        #   Tinta will ignore the separator
        assert (
            Tinta().purple(s).pink(s).to_str()
            == f"{expected_purple}{sep}{expected_pink}"
        )
        Tinta().purple(s).pink(s).print()
        with skip_on_github_actions() as skip:
            if not skip:
                out = capfd.readouterr().out
                assert out == f"{expected_purple}{sep}{expected_pink}\n"

    @pytest.mark.parametrize(
        "Testa,expected",
        whitespace_cases,
    )
    def test_proxy_color_chain_to_plaintext_respects_whitespace(
        self, Testa: Callable[[], Tinta], expected: str, capfd: CaptureFixture[str]
    ):
        p = Testa().to_str(plaintext=True)
        l_or_r_newline = expected.startswith("\n") or expected.endswith("\n")
        sep = "" if l_or_r_newline else " "
        # ^ If there is a newline at the end of the current or start of the next segment,
        #   Tinta will ignore the separator
        assert (
            Tinta().purple(p).pink(p).to_str(plaintext=True)
            == f"{expected}{sep}{expected}"
        )
        Tinta().purple(p).pink(p).print(plaintext=True)
        with skip_on_github_actions() as skip:
            if not skip:
                out = capfd.readouterr().out
                assert out == f"{expected}{sep}{expected}\n"

    def test_sep_inherits_prev_part_color(self):
        assert (
            Tinta().red("Red").green("Green").blue("Blue").to_str(sep=";")
            == f"{RED}Red;{GREEN}Green;{BLUE}Blue{O}"
        )


class TestMultipleInstances:

    def test_same_instance_chains_to_str(self):
        t = Tinta()
        red = f"{RED}Red"
        green = f"{GREEN}Green"
        blue = f"{BLUE}Blue"
        assert t.red("Red").to_str() == f"{red}{O}"
        assert t.green("Green").to_str() == f"{red} {green}{O}"
        assert t.blue("Blue").to_str() == f"{red} {green} {blue}{O}"

    def test_same_instance_resets_after_print(self):
        t = Tinta()
        assert t.red("Red").to_str() == f"{RED}Red{O}"
        t.print()
        assert t.green("Green").to_str() == f"{GREEN}Green{O}"
        t.print()
        assert t.blue("Blue").to_str() == f"{BLUE}Blue{O}"

    def test_multiple_instances_dont_interfere(self):
        t1 = Tinta()
        t2 = Tinta()
        t3 = Tinta()
        assert t1.red("Red").to_str() == f"{RED}Red{O}"
        assert t2.green("Green").to_str() == f"{GREEN}Green{O}"
        assert t3.blue("Blue").to_str() == f"{BLUE}Blue{O}"


# @pytest.mark.skip
class TestFeatures:

    def test_inspect(self):
        assert Tinta().inspect(code=0) == "default"
        assert Tinta().inspect(name="default") == 0

        assert Tinta().inspect(code=35) == "green"
        assert Tinta().inspect(name="green") == 35

    @pytest.mark.parametrize(
        "Testa,expected",
        [
            (lambda: Tinta("Hello")("world"), "Hello world"),
            (lambda: Tinta("Hello ")("world"), "Hello  world"),
            (lambda: Tinta("Hello")(", world"), "Hello, world"),
            (lambda: Tinta("Hello,")("world"), "Hello, world"),
            (lambda: Tinta("Hello")(",")("world"), "Hello, world"),
            (lambda: Tinta("Hello world")("!"), "Hello world!"),
            (lambda: Tinta("Nice")("."), "Nice."),
            (lambda: Tinta("Nice")(".gif file"), "Nice .gif file"),
            (lambda: Tinta(" *** Important"), " *** Important"),
            (lambda: Tinta(", world"), ", world"),
            (lambda: Tinta(" , world"), " , world"),
            (lambda: Tinta(",")("world"), ", world"),
            (lambda: Tinta(" ,")("world"), " , world"),
            (lambda: Tinta(" ,")(" world"), " ,  world"),
            (
                lambda: Tinta()
                .red(" *** Important : Smart fix punctuation should work with")
                .green("ANSI")
                .red(", as well as plaintext."),
                f"{RED} *** Important : Smart fix punctuation should work with {GREEN}ANSI{RED}, as well as plaintext.{O}",
            ),
        ],
    )
    def test_smart_fix_punctuation(self, Testa: Callable[[], Tinta], expected: str):
        assert Testa().to_str() == expected


class TestComplexStructure:

    @pytest.mark.parametrize(
        "inputs,expected",
        [
            (
                (
                    [
                        ("How long", None),
                        ("can two people talk about nothing?", None),
                        ("Hmm?", None),
                    ],
                    "How long can two people talk about nothing? Hmm?",
                )
            ),
            (
                (
                    [
                        ("How long", ""),
                        ("can two people talk about nothing?", ""),
                        ("Hmm?", ""),
                    ],
                    "How longcan two people talk about nothing?Hmm?",
                )
            ),
            (
                (
                    [
                        ("How long", " "),
                        ("can two people", " "),
                        ("talk about nothing?", " "),
                    ],
                    "How long can two people talk about nothing?",
                )
            ),
            (
                (
                    [("A section", " "), ("of text", ""), (", separated.", " ")],
                    "A section of text, separated.",
                )
            ),
        ],
    )
    def test_sep(self, inputs: List[Tuple[str, str]], expected: str) -> None:
        t = Tinta()
        for s, sep in inputs:
            t.push(s, sep=sep)
        assert t.to_str(plaintext=True) == expected

    def testparts(self):
        assert len(Tinta().green("green").red("red").blue("blue").parts) == 3
        assert (
            len(
                Tinta()
                .green("green")
                .underline()
                .red("red")
                .blue("blue")
                .yellow("yellow")
                .parts_fmt
            )
            == 4
        )
        assert (
            len(Tinta().green("green").underline().red("red").blue("blue").parts_pln)
            == 3
        )

    def test_f_strings(self):
        dog = "cat"
        assert (
            Tinta(f"A {dog} is a human's best friend").to_str(plaintext=True)
            == "A cat is a human's best friend"
        )
        assert Tinta(f"A {Tinta().red('hologram').to_str()} is a human's best friend")

    def test_push(self):
        t = Tinta().push("How long").push("can two people talk about nothing?")
        assert t.to_str(plaintext=True) == "How long can two people talk about nothing?"

        assert len(t.parts) == 2
        assert len(t.parts_fmt) == 2
        assert len(t.parts_pln) == 2

        t = Tinta().push("How long", "can two people talk about nothing?")
        assert t.to_str(plaintext=True) == "How long can two people talk about nothing?"

        assert len(t.parts) == 1
        assert len(t.parts_fmt) == 1
        assert len(t.parts_pln) == 1

    @pytest.mark.parametrize(
        "inputs,expected",
        [
            (
                ("How long", "can two people talk about nothing?"),
                "How long can two people talk about nothing?",
            ),
            (
                (["How long", "can two people talk about nothing?"]),
                "How long can two people talk about nothing?",
            ),
        ],
    )
    def test_to_str(self, inputs: Tuple[str], expected):
        t = Tinta().push(*inputs)
        assert t.to_str(plaintext=True) == expected
        assert t.to_str(plaintext=True) == expected

    def test_pop(self):
        t = Tinta().push("How long").push("can two people talk about nothing?")
        t.pop()
        assert t.to_str(plaintext=True) == "How long"
        assert len(t.parts) == 1
        assert len(t.parts_fmt) == 1
        assert len(t.parts_pln) == 1

        t = Tinta()
        for p in range(10):
            t.push(str(p))

        assert len(t.parts) == 10
        t.pop(10)

        assert len(t.parts) == 0

        t.pop(2)  # Shouldn't error if we remove more than we have

        assert len(t.parts) == 0

    def test_zero_handles_clear(self):
        s = (
            Tinta("White")
            .red("Red")
            .green("Green")
            .blue("Blue")
            .tint(0, "Clear")
            .to_str(sep=" ")
        )
        print(s)
        assert s == f"White {RED}Red {GREEN}Green {BLUE}Blue {O}Clear{O}"


class TestUtils:

    @pytest.mark.parametrize(
        "Testa,expected",
        [
            # fmt: off
            (lambda: Tinta().blue("blue"), "blue"),
            (lambda: Tinta().red("red"), "red"),
            (lambda: Tinta().green("green"), "green"),
            (lambda: Tinta().yellow("yellow"), "yellow"),
            (lambda: Tinta().pink("🦄"), "🦄"),
            # fmt: on
            (
                lambda: Tinta()
                .pink("Pink unicorns 🦄")
                .purple("are as good as purple 🦄 ones"),
                "Pink unicorns 🦄 are as good as purple 🦄 ones",
            ),
            (
                lambda: Tinta()
                .purple("Hello purple world!")
                .clear("I've got a song to sing about")
                ._("unicorns 🦄"),
                "Hello purple world! I've got a song to sing about unicorns 🦄",
            ),
        ],
    )
    def test_strip_ansi(self, Testa: Callable, expected: str):
        assert Tinta.strip_ansi(Testa().to_str()) == expected

    @pytest.mark.parametrize(
        "Testa, ljust, fillchar, expected",
        [
            # fmt: off
            (lambda: Tinta().mint("🦄"), 10, " ", f"{MINT}🦄{O}         "),
            (lambda: Tinta().mint("I love unicorns and magical fruit"), 10, " ", f"{MINT}I love unicorns and magical fruit{O}"),
            (lambda: Tinta("ljust"), 13, "-", f"ljust{'-'*8}"),
            # fmt: on
            (
                lambda: Tinta("Wild")
                .purple("unicorns")
                .pink("like their")
                .mint("space"),
                20,
                " ",
                f"Wild {PURPLE}unicorns {PINK}like their {MINT}space{O}",
            ),
            (
                lambda: Tinta("a").purple("b").pink("c d").mint("e "),
                80,
                " ",
                f"a {PURPLE}b {PINK}c d {MINT}e{O} {' '*70}",
            ),
        ],
    )
    def test_ljust(self, Testa: Callable, ljust: int, fillchar: str, expected: str):
        assert Tinta.ljust(Testa().to_str(), ljust, fillchar) == expected

    @pytest.mark.parametrize(
        "Testa, rjust, fillchar, expected",
        [
            # fmt: off
            (lambda: Tinta().mint("🦄"), 10, " ", f"         {MINT}🦄{O}"),
            (lambda: Tinta().mint("I love unicorns and magical fruit"), 10, " ", f"{MINT}I love unicorns and magical fruit{O}"),
            (lambda: Tinta("rjust"), 13, "-", f"{'-'*8}rjust"),
            # fmt: on
            (
                lambda: Tinta("Wild")
                .purple("unicorns")
                .pink("like their")
                .mint("space"),
                20,
                " ",
                f"Wild {PURPLE}unicorns {PINK}like their {MINT}space{O}",
            ),
            (
                lambda: Tinta(" a").purple("b").pink("c d").mint("e"),
                80,
                " ",
                f"{' '*70} a {PURPLE}b {PINK}c d {MINT}e{O}",
            ),
        ],
    )
    def test_rjust(self, Testa: Callable, rjust: int, fillchar: str, expected: str):
        assert Tinta.rjust(Testa().to_str(), rjust, fillchar) == expected



import sys
from typing import Any, Callable

# Literal
if sys.version_info >= (3, 8):
    from typing import Literal
elif sys.version_info >= (3, 5):
    from typing_extensions import Literal
else:
    from typing import Literal

# TypeVar
if sys.version_info >= (3, 5):
    from typing import TypeVar

    T = TypeVar("T")

# ParamSpec, GenericCallable
if sys.version_info >= (3, 10):
    from typing import Literal, ParamSpec

    P = ParamSpec("P")
    R = TypeVar("R")  # type: ignore
    GenericCallable = Callable[P, R]
else:
    GenericCallable = Any

# Deprecated
try:
    from warnings import deprecated
except ImportError:
    from deprecated import deprecated
deprecated = deprecated

all = [
    deprecated,
    GenericCallable,
    Literal,
    T,
    TypeVar,
]



#!/usr/bin/env python

"""
# Tinta
# Copyright 2024 github.com/brandoncript

# This program is bound to the Hippocratic License 2.1
# Full text is available here:
# https://firstdonoharm.dev/version/2/1/license

# Further to adherence to the Hippocratic License, permission is hereby
# granted, free of charge, to any person obtaining a copy of this software
# and associated documentation files (the "Software") under the terms of the
# MIT License to deal in the Software without restriction, including without
# limitation the rights to use, copy, modify, merge, publish, distribute,
# sublicense, and / or sell copies of the Software, and to permit persons
# to whom the Software is furnished to do so, subject to the conditions layed
# out in the MIT License.

# Where a conflict or dispute would arise between these two licenses, HLv2.1
# shall take precedence.
"""

from pathlib import Path

from setuptools import setup

with Path("README.md").open(encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="tinta",
    version="0.1.7b4-post0",
    description="Tinta, a magical console output tool.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/brandonscript/tinta",
    author="Brandon Shelley",
    author_email="brandon@pacificaviator.co",
    install_requires=[],
    include_package_data=True,
    license="MIT",
    packages=["tinta"],
    keywords="console colors ansi print terminal development",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Environment :: Console",
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
    zip_safe=False,
)
